import numpy as np
import dask.array as da
from dask_stitch.stitch import weight_block, merge_overlaps


def position_grid(shape, blocksize):
    """
    """

    coords = da.meshgrid(*[range(x) for x in shape], indexing='ij')
    coords = da.stack(coords, axis=-1).astype(np.int16)
    return da.rechunk(coords, chunks=tuple(blocksize) + (3,))


def affine_to_field(matrix, grid, displacement=True):
    """
    """

    # apply affine to coordinates
    mm = matrix[:3, :3]
    tt = matrix[:3, -1]
    result = np.einsum('...ij,...j->...i', mm, grid) + tt

    # convert positions to displacements
    if displacement:
        result = result - grid

    # return
    return result


def local_affines_to_field(
    shape, spacing, affines,
    blocksize, overlap,
    displacement=True,
):
    """
    """

    # get a coordinate grid
    grid = position_grid(
        np.array(blocksize) * affines.shape[:3], blocksize,
    )
    grid = grid * spacing.astype(np.float32)
    grid = grid[..., None]  # needed for map_overlap

    # wrap local_affines as dask array
    affines_da = da.from_array(
        affines, chunks=(1, 1, 1, 4, 4),
    )

    # convert to weighted fields
    def affine_to_weighted_field(x, y, block_info=None):
        field = affine_to_field(
            x.squeeze(), y.squeeze(),
            displacement=displacement,
        )
        return weight_block(
            field, blocksize, overlap, block_info,
        )

    # compute affine transforms as displacement fields, lazy dask arrays
    blocksize_with_overlaps = tuple(x+2*y for x, y in zip(blocksize, overlap))
    fields = da.map_overlap(
        affine_to_weighted_field,
        affines_da, grid,
        depth=[0, tuple(overlap)+(0, 0)],
        boundary=0,
        trim=False,
        align_arrays=False,
        dtype=np.float32,
        drop_axis=[4,],
        chunks=blocksize_with_overlaps+(3,),
    )

    # stitch affine position fields
    coords = da.map_overlap(
        merge_overlaps,
        fields,
        overlap=overlap,
        depth=tuple(overlap) + (0,),
        boundary=0.,
        trim=False,
        dtype=np.float32,
        chunks=list(blocksize) + [3,],
    )

    # crop to original shape
    coords = coords[:shape[0], :shape[1], :shape[2]]

    # return result
    return coords




import numpy as np
import dask.array as da
import copy
from itertools import product


def weight_block(block, blocksize, overlap, block_info=None):
    """
    """

    # determine which faces need linear weighting
    core, pad_ones, pad_linear = [], [], []
    block_index = block_info[0]['chunk-location']
    block_grid = block_info[0]['num-chunks']
    for i in range(3):

        # get core shape and pad sizes
        o = overlap[i]  # shorthand
        c = blocksize[i] - 2*o + 2
        p_ones, p_linear = [0, 0], [2*o-1, 2*o-1]
        if block_index[i] == 0:
            p_ones[0], p_linear[0] = 2*o-1, 0
        if block_index[i] == block_grid[i] - 1:
            p_ones[1], p_linear[1] = 2*o-1, 0
        core.append(c)
        pad_ones.append(tuple(p_ones))
        pad_linear.append(tuple(p_linear))

    # create weights core
    weights = np.ones(core, dtype=np.float32)

    # extend weights
    weights = np.pad(
        weights, pad_ones, mode='constant', constant_values=1,
    )
    weights = np.pad(
        weights, pad_linear, mode='linear_ramp', end_values=0,
    )

    # block may be a vector field
    # conditional is too general, but works for now
    if weights.ndim != block.ndim:
        weights = weights[..., None]

    # multiply data by weights and return
    return np.multiply(block, weights)


def merge_overlaps(block, overlap):
    """
    """

    o = overlap  # shorthand
    core = [slice(2*x, -2*x) for x in o]
    result = np.require(block[tuple(core)], requirements='W')

    # faces
    for ax in range(3):
        # the left side
        slc1 = [slice(None, None)]*3
        slc1[ax] = slice(0, o[ax])
        slc2 = copy.deepcopy(core)
        slc2[ax] = slice(0, o[ax])
        result[tuple(slc1)] += block[tuple(slc2)]
        # the right side
        slc1 = [slice(None, None)]*3
        slc1[ax] = slice(-1*o[ax], None)
        slc2 = copy.deepcopy(core)
        slc2[ax] = slice(-1*o[ax], None)
        result[tuple(slc1)] += block[tuple(slc2)]

    # edges
    for edge in product([0, 1], repeat=2):
        for ax in range(3):
            oo = np.delete(o, ax)
            left = [slice(None, oe) for oe in oo]
            right = [slice(-1*oe, None) for oe in oo]
            slc1 = [l if e == 0 else r for l, r, e in zip(left, right, edge)]
            slc2 = copy.deepcopy(slc1)
            slc1.insert(ax, slice(None, None))
            slc2.insert(ax, core[ax])
            result[tuple(slc1)] += block[tuple(slc2)]

    # corners
    for corner in product([0, 1], repeat=3):
        left = [slice(None, oe) for oe in o]
        right = [slice(-1*oe, None) for oe in o]
        slc = [l if c == 0 else r for l, r, c in zip(left, right, corner)]
        result[tuple(slc)] += block[tuple(slc)]

    return result


def stitch_blocks(blocks, blocksize, overlap):
    """
    """

    # blocks may be a vector fields
    # conditional is too general, but works for now
    if blocks.ndim != len(blocksize):
        blocksize = list(blocksize) + [3,]
        overlap = tuple(overlap) + (0,)

    # weight block edges
    weighted_blocks = da.map_blocks(
        weight_block, blocks,
        blocksize=blocksize[:3],
        overlap=overlap[:3],
        dtype=np.float32,
    )

    # stitch overlap regions, return
    return da.map_overlap(
        merge_overlaps, weighted_blocks,
        overlap=overlap[:3],
        depth=overlap,
        boundary=0.,
        trim=False,
        dtype=np.float32,
        chunks=blocksize,
    )

